import sys
import socket
import struct
from ctypes.util import find_library
from ctypes import Structure, pointer, c_long, CDLL
from poyonga.result import GroongaResult, GroongaSelectResult
from poyonga.const import GQTP_HEADER_SIZE

if sys.version_info[0] == 3:
    from urllib.request import urlopen
    from urllib.error import HTTPError
    from urllib.parse import urlencode
else:
    from urllib2 import urlopen, HTTPError
    from urllib import urlencode


def get_send_data_for_gqtp(cmd, **kwargs):
    """create cmd & send data to groonga"""
    _cmd = cmd
    _cmd_arg = "".join([" --%s '%s'" % (d, str(kwargs[d]).replace("'", r"\'")) for d in kwargs])
    _cmd = _cmd + _cmd_arg
    if sys.version_info[0] == 3:
        size = struct.pack("!I", len(_cmd.encode()))
        _header = b"".join([b"\xc7", b"\x00" * 7, size, b"\x00" * 12])
        _send_data = _header + _cmd.encode()
    else:
        size = struct.pack("!I", len(_cmd))
        _header = "".join(["\xc7", "\x00" * 7, size, "\x00" * 12])
        _send_data = _header + _cmd
    return _send_data


def convert_gqtp_result_data(_start, _end, status, raw_data):
    # struct result data
    diff_time = (_end.tv_sec + _end.tv_nsec / 1000000000.0) - (
        _start.tv_sec + _start.tv_nsec / 1000000000.0
    )
    if status != 0:
        status -= 65536
        body = '"",[["","",0]]'
        _data = "[[%d,%d.%d,%lf,%s]]" % (status, _start.tv_sec, _start.tv_nsec, diff_time, body)
    else:
        body = raw_data[GQTP_HEADER_SIZE:]
        if sys.version_info[0] == 3:
            body = body.decode()
        _data = "[[%d,%d.%d,%lf],%s]" % (status, _start.tv_sec, _start.tv_nsec, diff_time, body)
    return _data


class Groonga:

    LIBRT = CDLL(find_library("rt"))
    LIBC = CDLL(find_library("c"))

    class CTimeSpec(Structure):
        _fields_ = [("tv_sec", c_long), ("tv_nsec", c_long)]

    class CTimeval(Structure):
        _fields_ = [("tv_sec", c_long), ("tv_usec", c_long)]

    class _TimeSpec:
        tv_sec = 0.0
        tv_nsec = 0.0

    def __init__(
        self, host="localhost", port=10041, protocol="http", encoding="utf-8", prefix_path="/d/"
    ):
        self.host = host
        self.port = port
        self.protocol = protocol
        self.encoding = encoding
        self.prefix_path = prefix_path

    def _usec2nsec(self, nsec):
        return nsec * (1000000000 / 1000000)

    def _clock_gettime(self):
        ret = self._TimeSpec()
        if hasattr(self.LIBRT, "clock_gettime"):
            timespec = self.CTimeSpec()
            # 0: CLOCK_REALTIME
            self.LIBRT.clock_gettime(0, pointer(timespec))
            ret.tv_sec = timespec.tv_sec
            ret.tv_nsec = timespec.tv_nsec
        else:  # MacOSX and other environment
            timespec = self.CTimeval()
            self.LIBC.gettimeofday(pointer(timespec), None)
            ret.tv_sec = timespec.tv_sec
            ret.tv_nsec = self._usec2nsec(timespec.tv_usec)
        return ret

    def _call_gqtp(self, cmd, **kwargs):
        # create socket & send data
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((self.host, self.port))
        _start = self._clock_gettime()
        s.send(get_send_data_for_gqtp(cmd, **kwargs))
        # recv groonga data
        raw_data = s.recv(8192)
        proto, qtype, keylen, level, flags, status, size, opaque, cas = struct.unpack(
            "!BBHBBHIIQ", raw_data[:GQTP_HEADER_SIZE]
        )
        while len(raw_data) < size + GQTP_HEADER_SIZE:
            raw_data += s.recv(8192)
        _end = self._clock_gettime()
        s.close()
        return convert_gqtp_result_data(_start, _end, status, raw_data)

    def _call_http(self, cmd, **kwargs):
        domain = [self.protocol, "://", self.host, ":", str(self.port), self.prefix_path]
        url = "".join(domain) + cmd
        if kwargs:
            url = "".join([url, "?", urlencode(kwargs)])
        try:
            _data = urlopen(url).read()
        except HTTPError as msg:
            _data = msg.read()
        return _data

    def call(self, cmd, **kwargs):
        output_type = kwargs.get("output_type")
        if not output_type:
            output_type = "json"
        if self.protocol == "http":
            ret = self._call_http(cmd, **kwargs)
        else:
            ret = self._call_gqtp(cmd, **kwargs)
        if cmd == "select":
            return GroongaSelectResult(ret, output_type, self.encoding)
        else:
            return GroongaResult(ret, output_type, self.encoding)



import csv

try:
    from cStringIO import StringIO
except ImportError:
    try:
        from StringIO import StringIO
    except ImportError:
        from io import StringIO
try:
    import orjson as json
except ImportError:
    import json
try:
    import msgpack
except ImportError:
    msgpack = None


class GroongaResult:
    def __init__(self, data, output_type="json", encoding="utf-8"):
        self.raw_result = data
        if output_type == "tsv":
            # TODO: not implement
            csv.reader(StringIO(data), delimiter="\t")
        elif output_type == "msgpack":
            if msgpack:
                _result = msgpack.unpackb(data)
            else:
                raise Exception("msgpack is not support")
        elif output_type == "json":
            if encoding == "utf-8":
                _result = json.loads(data)
            else:
                _result = json.loads(data, encoding=encoding)
        else:  # xml or other types...
            # TODO: not implement
            pass
        self.status = _result[0][0]
        self.start_time = _result[0][1]
        self.elapsed = _result[0][2]
        if len(_result) == 1 and self.status != 0:
            self.body = _result[0][3]
        elif len(_result) == 1 and self.status == 0:
            self.body = ""  # handling invalid result
        else:
            self.body = _result[1]

    def __str__(self):
        return f"<{type(self).__name__} " + \
            f"status={self.status} " + \
            f"start_time={self.start_time} " + \
            f"elapsed={self.elapsed}>"


class GroongaSelectResult(GroongaResult):
    def __init__(self, data, output_type="json", encoding="utf-8"):
        super(GroongaSelectResult, self).__init__(data, output_type, encoding)
        self.items = []
        self.hit_num = -1  # default is -1 (Error)
        if len(self.body) != 0:
            self.hit_num = self.body[0][0][0]
        if self.status == 0:
            keys = [k[0] for k in self.body[0][1]]
            self.items = [dict(list(zip(keys, item))) for item in self.body[0][2:]]



# protocol type
GRN_PROTO_GQTP = "gqtp"
GRN_PROTO_HTTP = "http"

# gqtp
GQTP_HEADER_SIZE = 24

# groonga status
GRN_STATUS_SUCCESS = 0
GRN_STATUS_END_OF_DATA = 1
GRN_STATUS_UNKNOWN_ERROR = 65535
GRN_STATUS_OPERATION_NOT_PERMITTED = 65534
GRN_STATUS_NO_SUCH_FILE_OR_DIRECTORY = 65533
GRN_STATUS_NO_SUCH_PROCESS = 65532
GRN_STATUS_INTERRUPTED_FUNCTION_CALL = 65531
GRN_STATUS_INPUT_OUTPUT_ERROR = 65530
GRN_STATUS_NO_SUCH_DEVICE_OR_ADDRESS = 65529
GRN_STATUS_ARG_LIST_TOO_LONG = 65528
GRN_STATUS_EXEC_FORMAT_ERROR = 65527
GRN_STATUS_BAD_FILE_DESCRIPTOR = 65526
GRN_STATUS_NO_CHILD_PROCESSES = 65525
GRN_STATUS_RESOURCE_TEMPORARILY_UNAVAILABLE = 65524
GRN_STATUS_NOT_ENOUGH_SPACE = 65523
GRN_STATUS_PERMISSION_DENIED = 65522
GRN_STATUS_BAD_ADDRESS = 65521
GRN_STATUS_RESOURCE_BUSY = 65520
GRN_STATUS_FILE_EXISTS = 65519
GRN_STATUS_IMPROPER_LINK = 65518
GRN_STATUS_NO_SUCH_DEVICE = 65517
GRN_STATUS_NOT_A_DIRECTORY = 65516
GRN_STATUS_IS_A_DIRECTORY = 65515
GRN_STATUS_INVALID_ARGUMENT = 65514
GRN_STATUS_TOO_MANY_OPEN_FILES_IN_SYSTEM = 65513
GRN_STATUS_TOO_MANY_OPEN_FILES = 65512
GRN_STATUS_INAPPROPRIATE_I_O_CONTROL_OPERATION = 65511
GRN_STATUS_FILE_TOO_LARGE = 65510
GRN_STATUS_NO_SPACE_LEFT_ON_DEVICE = 65509
GRN_STATUS_INVALID_SEEK = 65508
GRN_STATUS_READ_ONLY_FILE_SYSTEM = 65507
GRN_STATUS_TOO_MANY_LINKS = 65506
GRN_STATUS_BROKEN_PIPE = 65505
GRN_STATUS_DOMAIN_ERROR = 65504
GRN_STATUS_RESULT_TOO_LARGE = 65503
GRN_STATUS_RESOURCE_DEADLOCK_AVOIDED = 65502
GRN_STATUS_NO_MEMORY_AVAILABLE = 65501
GRN_STATUS_FILENAME_TOO_LONG = 65500
GRN_STATUS_NO_LOCKS_AVAILABLE = 65499
GRN_STATUS_FUNCTION_NOT_IMPLEMENTED = 65498
GRN_STATUS_DIRECTORY_NOT_EMPTY = 65497
GRN_STATUS_ILLEGAL_BYTE_SEQUENCE = 65496
GRN_STATUS_SOCKET_NOT_INITIALIZED = 65495
GRN_STATUS_OPERATION_WOULD_BLOCK = 65494
GRN_STATUS_ADDRESS_IS_NOT_AVAILABLE = 65493
GRN_STATUS_NETWORK_IS_DOWN = 65492
GRN_STATUS_NO_BUFFER = 65491
GRN_STATUS_SOCKET_IS_ALREADY_CONNECTED = 65490
GRN_STATUS_SOCKET_IS_NOT_CONNECTED = 65489
GRN_STATUS_SOCKET_IS_ALREADY_SHUTDOWNED = 65488
GRN_STATUS_OPERATION_TIMEOUT = 65487
GRN_STATUS_CONNECTION_REFUSED = 65486
GRN_STATUS_RANGE_ERROR = 65485
GRN_STATUS_TOKENIZER_ERROR = 65484
GRN_STATUS_FILE_CORRUPT = 65483
GRN_STATUS_INVALID_FORMAT = 65482
GRN_STATUS_OBJECT_CORRUPT = 65481
GRN_STATUS_TOO_MANY_SYMBOLIC_LINKS = 65480
GRN_STATUS_NOT_SOCKET = 65479
GRN_STATUS_OPERATION_NOT_SUPPORTED = 65478
GRN_STATUS_ADDRESS_IS_IN_USE = 65477
GRN_STATUS_ZLIB_ERROR = 65476
GRN_STATUS_LZO_ERROR = 65475
GRN_STATUS_STACK_OVER_FLOW = 65474
GRN_STATUS_SYNTAX_ERROR = 65473
GRN_STATUS_RETRY_MAX = 65472
GRN_STATUS_INCOMPATIBLE_FILE_FORMAT = 65471
GRN_STATUS_UPDATE_NOT_ALLOWED = 65470
GRN_STATUS_TOO_SMALL_OFFSET = 65469
GRN_STATUS_TOO_LARGE_OFFSET = 65468
GRN_STATUS_TOO_SMALL_LIMIT = 65467
GRN_STATUS_CAS_ERROR = 65466
GRN_STATUS_UNSUPPORTED_COMMAND_VERSION = 65465



# coding: utf-8
import json
import sys
import struct
import unittest

try:
    from mock import patch, Mock
except ImportError:
    from unittest.mock import patch, Mock
from poyonga import Groonga, GroongaResult
from poyonga.client import get_send_data_for_gqtp, convert_gqtp_result_data, GQTP_HEADER_SIZE
from poyonga.const import GRN_STATUS_UNSUPPORTED_COMMAND_VERSION


class PoyongaHTTPTestCase(unittest.TestCase):
    def setUp(self):
        self.g = Groonga()

    @patch("poyonga.client.urlopen")
    def test_json_result_with_http(self, mock_urlopen):
        m = Mock()
        m.read.side_effect = ["[[0, 1337566253.89858, 0.000354], {}]"]
        mock_urlopen.return_value = m
        ret = self.g.call("status")
        self.assertEqual(type(ret), GroongaResult)
        self.assertEqual(ret.status, 0)


class PoyongaGQTPTestCase(unittest.TestCase):
    def setUp(self):
        self.g = Groonga(protocol="gqtp")

    @patch("poyonga.client.socket.socket")
    def test_json_result_with_gqtp(self, mock_socket):
        m = Mock()
        if sys.version_info[0] == 3:
            _data = b"{}"
        else:
            _data = "{}"
        _proto, _qtype, _keylen, _level, _flags, _status, _size, _opaque, _cas = (
            0xC7,
            0x02,
            0,
            0,
            0,
            0,
            2,
            0,
            0,
        )
        packdata = struct.pack(
            "!BBHBBHIIQ2s",
            _proto,
            _qtype,
            _keylen,
            _level,
            _flags,
            _status,
            _size,
            _opaque,
            _cas,
            _data,
        )
        m.recv.return_value = packdata
        mock_socket.return_value = m
        ret = self.g.call("status")
        self.assertEqual(type(ret), GroongaResult)
        self.assertEqual(ret.status, 0)


class PoyongaFunctions(unittest.TestCase):
    def test_get_send_data(self):
        d = get_send_data_for_gqtp("status")
        self.assertEqual(30, len(d))
        if sys.version_info[0] == 3:
            self.assertEqual(ord("\xc7"), d[0])
        else:
            self.assertEqual("\xc7", d[0])
        (size,) = struct.unpack("!I", d[8:12])
        self.assertEqual(6, size)

    def test_get_send_data_with_args(self):
        kwargs = {"table": "Site"}
        d = get_send_data_for_gqtp("select", **kwargs)
        self.assertEqual(45, len(d))
        if sys.version_info[0] == 3:
            self.assertEqual(ord("\xc7"), d[0])
            # check body length
            self.assertEqual(21, d[11])
        else:
            self.assertEqual("\xc7", d[0])
            # check body length
            self.assertEqual("\x15", d[11])

        body = d[GQTP_HEADER_SIZE:]
        (size,) = struct.unpack("!I", d[8:12])
        self.assertEqual(len(body), size)

    def test_get_send_data_with_args_and_two_bytes(self):
        kwargs = {"table": "Sitは"}
        d = get_send_data_for_gqtp("select", **kwargs)
        self.assertEqual(47, len(d))
        if sys.version_info[0] == 3:
            self.assertEqual(ord("\xc7"), d[0])
            # check body length
            self.assertEqual(23, d[11])
        else:
            self.assertEqual("\xc7", d[0])
            # check body length
            self.assertEqual("\x17", d[11])

        body = d[GQTP_HEADER_SIZE:]
        (size,) = struct.unpack("!I", d[8:12])
        self.assertEqual(len(body), size)

    def test_convert_gqtp_result(self):
        grn = Groonga()
        s = grn._clock_gettime()
        e = grn._clock_gettime()
        senddata = get_send_data_for_gqtp("status")
        rawdata = senddata[:GQTP_HEADER_SIZE] + b'{"test": 0}'
        d = convert_gqtp_result_data(s, e, 0, rawdata)
        d = json.loads(d)
        self.assertEqual(d[0][0], 0)

    def test_convert_gqtp_result_fail(self):
        grn = Groonga()
        s = grn._clock_gettime()
        e = grn._clock_gettime()
        senddata = get_send_data_for_gqtp("status")
        rawdata = senddata[:GQTP_HEADER_SIZE] + b'{"test": 0}'
        d = convert_gqtp_result_data(s, e, GRN_STATUS_UNSUPPORTED_COMMAND_VERSION, rawdata)
        d = json.loads(d)
        self.assertEqual(d[0][0], -71)


class PoyongaResult(unittest.TestCase):
    def test_str(self):
        status = 0
        start_time = 0.0
        elapsed = 1.0
        raw_result = [[status, start_time, elapsed], False]
        result = GroongaResult(json.dumps(raw_result))
        self.assertEqual('<GroongaResult ' +
                         f'status={status} ' +
                         f'start_time={start_time} ' +
                         f'elapsed={elapsed}>',
                         str(result))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
import poyonga

setup(
    name="poyonga",
    version=poyonga.__version__,
    description="Python Groonga Client",
    long_description=open("README.rst").read(),
    license="MIT License",
    author="Hideo Hattori",
    author_email="hhatto.jp@gmail.com",
    url="https://github.com/hhatto/poyonga",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: Unix",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    keywords="groonga http gqtp",
    packages=["poyonga"],
    zip_safe=False,
)
